package org.openxma.dsl.generator.component;

import java.io.File;

import org.eclipse.core.resources.IFolder;
import org.eclipse.core.resources.IWorkspaceRoot;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.Path;
import org.eclipse.emf.common.util.URI;
import org.eclipse.xpand2.output.FileHandle;
import org.eclipse.xpand2.output.PostProcessor;
import org.openxma.dsl.common.logging.LogUtil;
import org.openxma.dsl.generator.GeneratorConfiguration;
import org.openxma.hbmfilemerge.HbmFileMerge;
import org.openxma.hbmfilemerge.HbmFileMergeInput;

public class HbmFileMergePostProcessor implements PostProcessor {
    private File outletResources, outletGeneratedResources;

    @SuppressWarnings("deprecation")
    public void afterClose(FileHandle impl) {
        initPath();
        File genHbm = impl.getTargetFile();

        // because the gen file is created after the normal hbm file we react to
        // the creation of gen hbm files
        // because then we can merge into the normal hbm files without loosing
        // the merge result
        if (!isExitistingHbmGenFileInGenResourcesFolder(genHbm))
            return;
        try {
            HbmFileMerge.getInstance(outletResources, outletGeneratedResources).mergeFragmentHbmWith(genHbm);
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    private void initPath() {
        if (null != outletResources && null != outletGeneratedResources) {
            return;
        }
        outletResources = getOutletFolder(GeneratorConfiguration.INSTANCE.getResourceFolder());
        outletGeneratedResources = getOutletFolder(GeneratorConfiguration.INSTANCE.getGenResourceFolder());
    }

    private File getOutletFolder(URI uri) {
        if (uri.isPlatformResource()) {
            IWorkspaceRoot root = ResourcesPlugin.getWorkspace().getRoot();
            IFolder folder = root.getFolder((IPath) toPath(uri));
            return new File(folder.getLocationURI());
        }
        return new File(GeneratorConfiguration.INSTANCE.getResourceFolder().toFileString());
    }

    private Object toPath(URI uri) {
        return new Path(uri.toPlatformString(true));
    }

    /**
     * @param delta
     *            , the delta of the changed resource
     * @return <code>true</code> if the changed recource/file has the file name suffix
     *         <code>HbmFileChangeListener.FRAGMENT_HBM_FILE_SUFFIX</code> and is within the Java Source folder
     */
    private boolean isExitistingHbmGenFileInGenResourcesFolder(File toCheck) {
        if (toCheck == null || !toCheck.exists() || !toCheck.isFile()) {
            LogUtil.info("file does not exist or is not a file: \n" + toCheck.getAbsolutePath());
            return false;
        }

        return (HbmFileMergeInput.isGenHbmFile(toCheck.getName()) && isWithinGenResourceFolder(toCheck)) ? true : false;
    }

    private boolean isWithinGenResourceFolder(File toCheck) {
        return toCheck.getAbsolutePath().startsWith(outletGeneratedResources.getAbsolutePath()) ? true : false;
    }

    public void beforeWriteAndClose(FileHandle impl) {
    }

}
