/*
 * @(#) $Id: InvokeStrategy.java 8808 2011-12-14 12:58:06Z laslovd $
 *
 * Copyright 2009/2010 by sIT Solutions,
 * A-1110 Wien, Geiselbergstr.21-25.
 * All rights reserved.
 *
 */
package org.openxma.dsl.generator.impl.substrategy;

import java.util.Set;

import org.eclipse.xtext.util.Strings;
import org.openxma.dsl.core.model.SimpleType;
import org.openxma.dsl.core.model.VariableAccess;
import org.openxma.dsl.generator.helper.BasicTypeHelper;
import org.openxma.dsl.generator.helper.WidgetType;
import org.openxma.dsl.generator.helper.style.BasicDataType;
import org.openxma.dsl.generator.impl.PageStatementInternalIterator;
import org.openxma.dsl.pom.model.CheckBox;
import org.openxma.dsl.pom.model.Combo;
import org.openxma.dsl.pom.model.Component;
import org.openxma.dsl.pom.model.CustomizeableField;
import org.openxma.dsl.pom.model.GuiElement;
import org.openxma.dsl.pom.model.IReferenceableByStatementVariable;
import org.openxma.dsl.pom.model.InvokeStatement;
import org.openxma.dsl.pom.model.Invokeable;
import org.openxma.dsl.pom.model.JoinDirection;
import org.openxma.dsl.pom.model.ObjectProperty;
import org.openxma.dsl.pom.model.Page;
import org.openxma.dsl.pom.model.PropertyJoin;
import org.openxma.dsl.pom.model.Proxy;
import org.openxma.dsl.pom.model.ReferencedXMAPage;
import org.openxma.dsl.pom.model.Statement;
import org.openxma.dsl.pom.model.StatementVariable;
import org.openxma.dsl.pom.model.Table;
import org.openxma.dsl.pom.model.Text;
import org.openxma.dsl.pom.model.XmadslPage;
import org.openxma.dsl.pom.naming.PomNameProvider;

import at.spardat.xma.guidesign.XMAPage;

public class InvokeStrategy {
    
    public static void genAdditionalImportsClient(XMAPage page, final Set<String> imports) {
        new PageStatementInternalIterator(page) {
            @Override
            protected boolean visit(Statement stmt) {
                if ( stmt instanceof InvokeStatement ) {
                    InvokeStatement invokeStmt = (InvokeStatement) stmt;
                    boolean isEmbedding = isEmbedding(invokeStmt);
                    boolean isComponent = invokeStmt.getInvokeAble() instanceof Component;
                    if ( isEmbedding || isComponent ) {
                        if ( isEmbedding ) { 
                            imports.add("java.util.HashMap");
                        }
                        imports.add("at.spardat.xma.boot.component.*");
                        if ( isComponent ) {
                            Component component = (Component) invokeStmt.getInvokeAble();
                            String pkg = component.getXmaComponent().getNamPackage() + ".client.";
                            String clazz = InvokeStrategy.getInvokableTypeName(component);
                            imports.add(pkg + clazz);
                        }
                    }
                }
                return false;
            }
        }.iterate();
    }

    public static void genAdditionalMemberVariablesClient(XMAPage page, final StringBuilder sb) {
        new PageStatementInternalIterator(page) {
            @Override
            protected boolean visit(Statement stmt) {
                if ( isEmbedding(stmt) ) {
                    sb.append("    private HashMap<Composite,IXMAControl> compositeToEmbedableMap = new HashMap<Composite,IXMAControl>();\n");
                    return true; // stop after first
                }
                return false;
            }
        }.iterate();
    }
    
    public static void genAdditionalMethodsClientbuf(final StringBuilder buf, XMAPage page) {
        new PageStatementInternalIterator(page) {
            @Override
            protected boolean visit(Statement stmt) {
                if ( isEmbedding(stmt) ) {
                    buf.append("    /* Embeds the given IXMAControl into the given composite */\n"
                              +"    private void embedEmbedabel(Composite composite, IXMAControl embedable) {\n"
                              +"        Composite compW = embedable.createComposite(composite);\n"
                              +"        FormData data = new FormData();\n"
                              +"        data.left = new FormAttachment(0, 0);\n"
                              +"        data.right = new FormAttachment(100, 0);\n"
                              +"        data.top = new FormAttachment(0, 0);\n"
                              +"        data.bottom = new FormAttachment(100, 0);\n"
                              +"        compW.setLayoutData(data);\n"
                              +"        composite.setTabList(new Control[] { compW });\n"
                              +"        addChild(embedable);\n"
                              +"        compositeToEmbedableMap.put(composite,embedable);\n"
                              +"    }\n"
                              +"\n"
                              +"    /* Un-Embeds an eventually embedded page or component */\n"
                              +"    private void unembedEmbedabel(Composite composite) {\n"
                              +"        IXMAControl oldEmbedable = (IXMAControl) compositeToEmbedableMap.get(composite);\n"
                              +"        if (oldEmbedable != null) {\n"
                              +"            removeChild(oldEmbedable);\n"
                              +"            if ( oldEmbedable instanceof IComponent ) {\n"
                              +"                ((IComponent)oldEmbedable).dispose();\n"
                              +"            }\n"
                              +"            compositeToEmbedableMap.put(composite,null);\n"
                              +"        }\n"
                              +"    }\n"
                              +"\n");
                    return true; // stop after first
                }
                return false;
            }
        }.iterate();
    }

    
    public static void genInvokeImplementationClient(XMAPage page, InvokeStatement stmt, StringBuilder sb) {
        String typeName = getInvokableTypeName(stmt.getInvokeAble());
        String invokeableName = Strings.toFirstLower(PomNameProvider.getXmaName(stmt.getInvokeAble()));                       
        if ( isEmbedding(stmt) ) {
            String targetPanelName = stmt.getTargetPanel().getName() + "W";
            sb.append("        unembedEmbedabel("+targetPanelName+");\n");

            if ( stmt.getInvokeAble() instanceof Page ) {
                sb.append("        "+typeName+" "+invokeableName+" = new "+typeName+"(this);\n");
                genInProperties(stmt,sb);
            }
            else if ( stmt.getInvokeAble() instanceof Component ) {
                sb.append("        "+typeName+" "+invokeableName+" = ("+typeName+")getComponent().getSession().getComponent("+typeName+".COMPONENT_NAME);\n");
                genInProperties(stmt,sb);
            }
            else if ( stmt.getInvokeAble() instanceof Proxy ) {
                String url = ((Proxy)stmt.getInvokeAble()).getUrl();
                sb.append("        "+typeName+" "+invokeableName+" = getComponent().getSession().getComponentExtern(\""+url+"\");\n");
                genInProperties(stmt,sb);
            }
            sb.append("        embedEmbedabel("+targetPanelName+",(IXMAControl)"+invokeableName+");\n");
        }
        else {
            if ( stmt.getInvokeAble() instanceof Page ) {
                sb.append("        "+typeName+" "+invokeableName+" = new "+typeName+"(this);\n");
                genInProperties(stmt,sb);
                sb.append("        boolean ok = "+invokeableName+".invoke();\n");
                sb.append("        if (ok) {\n");
                genOutProperties(stmt,sb);
                sb.append("        }\n");
            }
            else {
                if ( stmt.getInvokeAble() instanceof Component ) {
                    sb.append("        "+typeName+" "+invokeableName+" = ("+typeName+")getComponent().getSession().getComponent("+typeName+".COMPONENT_NAME);\n");
                }
                else if ( stmt.getInvokeAble() instanceof Proxy ) {
                    String url = ((Proxy)stmt.getInvokeAble()).getUrl();
                    sb.append("        IComponent "+invokeableName+" = getComponent().getSession().getComponentExtern(\""+url+"\");\n");
                }
                sb.append("        try {\n");
                genInProperties("    ",stmt,sb);
                sb.append("            "+invokeableName+".invoke(getDialog().getShell());\n");
                genOutProperties(stmt,sb);
                sb.append("        } finally {\n");
                sb.append("            "+invokeableName+".dispose();\n");
                sb.append("        }\n");
            }
        }
    }

    private static void genInProperties(InvokeStatement stmt, StringBuilder sb) {
        genInProperties(null,stmt,sb);
    }           
    
    private static void genInProperties(String additionalIndent, InvokeStatement stmt, StringBuilder sb) {
        for (PropertyJoin propertyJoin : stmt.getPropertyJoin()) {
            if (propertyJoin.getJoinDirection() == JoinDirection.IN || propertyJoin.getJoinDirection() == JoinDirection.INOUT ) {
               
                /* 1. generate root */
                sb.append("        ");
                if (additionalIndent != null) {
                    sb.append(additionalIndent);
                }
                String xmaName = Strings.toFirstLower(PomNameProvider.getXmaName(stmt.getInvokeAble()));                
                sb.append(xmaName+".");                
                
                /* 2. determine types */
                InvokePropertyType sourceType = null;                    
                StatementVariable var = propertyJoin.getVariable();
                if (var != null) {
                    sourceType = getPropertyJoinValueType(var.getReference(),propertyJoin.getAccess());            
                } else if (propertyJoin.getStringLiteral() != null) {
                    sourceType = new InvokePropertyType(InvokePropertyTypEnum.String);
                } else {
                    // because no other value is defined
                    // TODO change model to allow explicit determination
                    sourceType = new InvokePropertyType(InvokePropertyTypEnum.Integer);
                    sourceType.setPrimitive(true);
                }
                InvokePropertyType targetType = getPropertyJoinValueType(propertyJoin.getProperty(),null);                 
                
                /* 4. value to set */
                String valueToSetString = generateInPropertyValue(propertyJoin,sourceType,targetType);                
                
                /* 5. converted value to set */
                String convertedValueToSetString = generateTypeConversion(sourceType,targetType,valueToSetString);
                
                /* 6. setter */
                String closeSetterString = generateInPropertySetter(propertyJoin,sourceType,targetType,sb);                
                sb.append(convertedValueToSetString);
                if (closeSetterString != null) {
                    sb.append(closeSetterString);
                }
                
                /* end of statement */
                sb.append(";\n");
            }
        }
    }
    
    private static void genOutProperties(InvokeStatement stmt, StringBuilder sb) {
        for (PropertyJoin propertyJoin : stmt.getPropertyJoin()) {
            if (propertyJoin.getJoinDirection() == JoinDirection.OUT || propertyJoin.getJoinDirection() == JoinDirection.INOUT ) {
               
                /* 1. generate root */
                sb.append("            ");
                String xmaName = Strings.toFirstLower(PomNameProvider.getXmaName(stmt.getInvokeAble()));                                    
                
                /* 2. determine types */
                InvokePropertyType sourceType = getPropertyJoinValueType(propertyJoin.getProperty(),null);                
                InvokePropertyType targetType = null;                    
                StatementVariable var = propertyJoin.getVariable();
                if (var != null) {
                    targetType = getPropertyJoinValueType(var.getReference(),propertyJoin.getAccess());            
                } else if (propertyJoin.getStringLiteral() != null) {
                    targetType = new InvokePropertyType(InvokePropertyTypEnum.String);
                } else {
                    // because no other value is defined
                    // TODO change model to allow explicit determination
                    targetType = new InvokePropertyType(InvokePropertyTypEnum.Integer);
                    targetType.setPrimitive(true);                    
                }                 
                
                /* 4. value to set */
                String valueToSetString = xmaName+"."+generateOutPropertyValue(propertyJoin,sourceType,targetType);
                
                /* 5. converted value to set */
                String convertedValueToSetString = generateTypeConversion(sourceType,targetType,valueToSetString);
                
                /* 6. setter */
                String closeSetterString = generateOutPropertySetter(propertyJoin,targetType,sourceType,sb);                
                sb.append(convertedValueToSetString);
                if (closeSetterString != null) {
                    sb.append(closeSetterString);
                }
                
                /* end of statement */
                sb.append(";\n");
            }
        }               
    }    
    
    private static InvokePropertyType getPropertyJoinValueType(IReferenceableByStatementVariable ref, VariableAccess variableAccess) {
        InvokePropertyType type = null;
        if (ref instanceof ObjectProperty) {
            ObjectProperty property = (ObjectProperty)ref;
            type = new InvokePropertyType(property.getType().getName());
            // Some types of component properties and widgetmodels are primitive
            switch (type.typeEnum) {
            case Boolean:            
            case String:
                break;
            case Integer:
            case Long:
            case Double:
                if (property.eContainer() instanceof Component) {
                    type.setPrimitive(true);
                }
                break;
//            default:
//                throw new RuntimeException("Unsupported property type: " + property.getType().getName());
            }                        
            return type;
        } else if (ref instanceof GuiElement) {
            GuiElement guiElement = (GuiElement)ref;
            BasicDataType basicDataType = null;
            if (guiElement instanceof CustomizeableField) { 
                basicDataType = BasicTypeHelper.determineBasicDataType(((CustomizeableField)ref));
            }            
            WidgetType widgetType = BasicTypeHelper.determineWidgetType(guiElement);
            switch(widgetType) {
            case TEXT:              
            if (   (variableAccess == null || variableAccess.getValue() == VariableAccess.NULL.getValue())
                || variableAccess.getValue() == VariableAccess.VALUE_VALUE                                 ) {
                    if (basicDataType != null) {
                        switch(basicDataType) {
                        case String:
                            return new InvokePropertyType(InvokePropertyTypEnum.String);
                        case Integer:
                            return new InvokePropertyType(InvokePropertyTypEnum.Integer);
                        case Long:   
                            return new InvokePropertyType(InvokePropertyTypEnum.Long);
                        case Double:   
                            return new InvokePropertyType(InvokePropertyTypEnum.Double);
                        case Float:   
                            return new InvokePropertyType(InvokePropertyTypEnum.Float);
                        case BigDecimal:   
                            return new InvokePropertyType(InvokePropertyTypEnum.BigDecimal);
                        case Date:   
                            return new InvokePropertyType(InvokePropertyTypEnum.Date);                              
                        }                        
                    } else {                    
                return new InvokePropertyType(InvokePropertyTypEnum.String);
                    }
            } else {
                if (variableAccess.getValue() == VariableAccess.IS_VALID_VALUE) {   
                        type = new InvokePropertyType(InvokePropertyTypEnum.Boolean);
                    type.setPrimitive(true);            
                    return type;                                    
                } else if (variableAccess.getValue() == VariableAccess.HAS_VALUE_VALUE) {
                        type = new InvokePropertyType(InvokePropertyTypEnum.Boolean);
                    type.setPrimitive(true);            
                    return type;
                } else {
                    throw new RuntimeException("Wdiget models of type 'Text' does not support : " + variableAccess.getLiteral());
                }
            } 
            case CHECKBOX:
                type = new InvokePropertyType(InvokePropertyTypEnum.Boolean);
            type.setPrimitive(true);            
            return type;
            case DATEPICKER:                
                return new InvokePropertyType(InvokePropertyTypEnum.Date);                
            case SIMPLECOMBO:
            case DOMAINCOMBO:                
            return new InvokePropertyType(InvokePropertyTypEnum.String);                        
            case TABLE:
                type = null;                
            if (    (variableAccess == null || variableAccess.getValue() == VariableAccess.NULL.getValue())
                 || variableAccess.getValue() == VariableAccess.GET_SELECTED_VALUE                          ) {
                return new InvokePropertyType(InvokePropertyTypEnum.String);
            } else if (variableAccess.getValue() == VariableAccess.GET_SELECTION_COUNT_VALUE) {
                       type = new InvokePropertyType(InvokePropertyTypEnum.Integer);
                type.setPrimitive(true);            
                return type;
            } else if (variableAccess.getValue() == VariableAccess.SIZE_VALUE) {
                       type = new InvokePropertyType(InvokePropertyTypEnum.Integer);
                type.setPrimitive(true);            
                return type;                                    
            } else {
                throw new RuntimeException("Wdiget models of type 'Table' does not support : " + variableAccess.getLiteral());
            }
            default:
                throw new RuntimeException("Widget type not supported: "+widgetType);
        }
        }                        
        throw new RuntimeException("Unable to dertermine type for: " + ref.toString());
    }
    
    private static String generateTypeConversion(InvokePropertyType sourceType, InvokePropertyType targetType, String valueToSetString) {                
        if (sourceType.typeEnum == targetType.typeEnum) {
            return valueToSetString;
        }
        switch(targetType.typeEnum) {
        case String:
            switch(sourceType.typeEnum) {
            case Long:
                if (sourceType.isPrimitive()) {
                    return "Long.toString("+valueToSetString+")";                    
                } else {
                    return valueToSetString+"!= null ? Long.toString("+valueToSetString+") : null";
                }
            case Integer:
                if (sourceType.isPrimitive()) {
                    return "Integer.toString("+valueToSetString+")";                    
                } else {
                    return valueToSetString+"!= null ? Integer.toString("+valueToSetString+") : null";
                }                                
            }
        case Long:
            switch(sourceType.typeEnum) {
            case String:
                if (targetType.isPrimitive()) {
                    return "("+valueToSetString+"!= null && "+valueToSetString+".length() > 0) ? Long.parseLong("+valueToSetString+") : 0";                    
                } else {
                    return "("+valueToSetString+"!= null && "+valueToSetString+".length() > 0) ? Long.parseLong("+valueToSetString+") : null";
                }
            case Integer:
                if (sourceType.isPrimitive() || targetType.isPrimitive()) {                
                    return "(long)"+valueToSetString;
                }                
            }
        case Integer:
            switch(sourceType.typeEnum) {
            case String:
                if (targetType.isPrimitive()) {
                    return "("+valueToSetString+"!= null && "+valueToSetString+".length() > 0) ? Integer.parseInt("+valueToSetString+") : 0";
                } else {
                    return "("+valueToSetString+"!= null && "+valueToSetString+".length() > 0) ? Integer.parseInt("+valueToSetString+") : null";
                }
            case Long:
                if (sourceType.isPrimitive() || targetType.isPrimitive()) {                
                    return "(int)"+valueToSetString;
                } 
            }              
        default: throw new RuntimeException("Type conversion for property join is not supported (source = '"+sourceType.typeEnum+"', target = '"+targetType.typeEnum+"')");
        }
    }        
    
    private static String generateInPropertyValue(PropertyJoin propertyJoin, InvokePropertyType sourceType, InvokePropertyType targetType) {
        StatementVariable var = propertyJoin.getVariable();
        if (var != null) {
            IReferenceableByStatementVariable ref = var.getReference();               
            String valueString = generatePropertyValue(ref,propertyJoin.getAccess(),sourceType,targetType);
            if (ref instanceof ObjectProperty) {
                ObjectProperty objectProperty = (ObjectProperty)ref;
                if (objectProperty.eContainer() instanceof Component) {
                    valueString="getTypedComponent()."+valueString;
                }
            }
            return valueString;
        } else if (propertyJoin.getStringLiteral() != null) {
            return "\""+propertyJoin.getStringLiteral()+"\"";
        } else {   
            return Integer.toString(propertyJoin.getIntegerLiteral());            
        }              
    }
    
    private static String generateOutPropertyValue(PropertyJoin propertyJoin, InvokePropertyType sourceType, InvokePropertyType targetType) {
        return generatePropertyValue(propertyJoin.getProperty(), null, sourceType,targetType);
    }    
    
    private static String generatePropertyValue(IReferenceableByStatementVariable ref, VariableAccess variableAccess, InvokePropertyType sourceType, InvokePropertyType targetType) {
        StringBuilder sb = new StringBuilder();      
        if (ref instanceof ObjectProperty) {
            ObjectProperty property = (ObjectProperty)ref;        
            if (property.eContainer() instanceof Component) {
                sb.append("get"+Strings.toFirstUpper(property.getName())+"()");           
            } else {
                sb.append(Strings.toFirstLower(property.getName())+".");
                switch(sourceType.typeEnum) {
                case String: 
                    sb.append("toString()");
                    break;
                case Long:
                    if (targetType.isPrimitive()) {
                        sb.append("toLong()");                        
                    } else {
                        sb.append("toLONG()");
                    }
                    break;
                case Integer:
                    if (targetType.isPrimitive()) {                    
                        sb.append("toInt()");                        
                    } else {
                        sb.append("toINTEGER()");
                    }
                    break;
                case Boolean:
                    sb.append("isTrue()");
                }            
            }
        } else if (ref instanceof GuiElement) {
            GuiElement guiElement = (GuiElement)ref;
            String modelName = Strings.toFirstLower(PomNameProvider.getXmaName(guiElement));
            WidgetType widgetType = BasicTypeHelper.determineWidgetType(guiElement);
            BasicDataType basicDataType = null;
            if (ref instanceof CustomizeableField) { 
                basicDataType = BasicTypeHelper.determineBasicDataType(((CustomizeableField)ref));
            }
            switch(widgetType) {
            case TEXT:              
            if (variableAccess != null && variableAccess.getValue() == VariableAccess.IS_VALID_VALUE) {            
                sb.append(modelName+".isValueValid()");                
            } else if (variableAccess != null && variableAccess.getValue() == VariableAccess.HAS_VALUE_VALUE) {            
                sb.append(modelName+".hasValue()");                
            } else {
                    if (basicDataType != null) {
                        switch(basicDataType) {
                        case String:
                            sb.append(modelName+".toString()");
                            break;
                        case Integer:
                            if (targetType.isPrimitive()) {                    
                                sb.append(modelName+".toInt()");                        
                            } else {
                                sb.append(modelName+".toINTEGER()");
                            }
                            break;   
                        case Long:   
                            if (targetType.isPrimitive()) {
                                sb.append(modelName+".toLong()");                        
                            } else {
                                sb.append(modelName+".toLONG()");
                            }
                            break;
                        case Double:   
                            if (targetType.isPrimitive()) {
                                sb.append(modelName+".toDouble()");                        
                            } else {
                                sb.append(modelName+".toDOUBLE()");
                            }   
                            break;
                        case Float:   
                            if (targetType.isPrimitive()) {
                                sb.append(modelName+".toFloat()");                        
                            } else {
                                sb.append(modelName+".toFLOAT()");
                            }   
                            break;
                        case BigDecimal:   
                            sb.append(modelName+".toBigDecimal()");
                            break;
                        case Date:   
                            sb.append(modelName+".toDate()");
                            break;                                                        
                        }
                        
                    } else {
                sb.append(modelName+".toString()");                
            }
                }  
                break;
            case CHECKBOX:
            sb.append(modelName+".isTrue()");            
                break;
            case DATEPICKER:
                sb.append(modelName+".toDate()");   
                break;                
            case SIMPLECOMBO:
            case DOMAINCOMBO:                
            sb.append(modelName+".getSelected()");            
                break;
            case TABLE:
            if (variableAccess != null && variableAccess.getValue() == VariableAccess.GET_SELECTION_COUNT_VALUE) {
                sb.append(modelName+".getSelectionCount()");
            } else if (variableAccess.getValue() == VariableAccess.SIZE_VALUE) {
                sb.append(modelName+".size()");                
            } else {
                sb.append(modelName+".getSelected()");
            }
                break;
            default:
                throw new RuntimeException("Widget type not supported: "+widgetType);
        }
        } 
        return sb.toString();
    }                
    
    private static String generateInPropertySetter(PropertyJoin propertyJoin, InvokePropertyType sourceType, InvokePropertyType targetType, StringBuilder sb) {
        return generatePropertySetter(propertyJoin.getProperty(),sourceType,targetType,sb);
    }
    
    private static String generateOutPropertySetter(PropertyJoin propertyJoin, InvokePropertyType sourceType, InvokePropertyType targetType, StringBuilder sb) {
        StatementVariable var = propertyJoin.getVariable();
        if (var != null) {
            IReferenceableByStatementVariable ref = var.getReference();              
            if (ref instanceof ObjectProperty) {
                ObjectProperty objectProperty = (ObjectProperty)ref;
                if (objectProperty.eContainer() instanceof Component) {
                    sb.append("getTypedComponent().");
                }
            } 
            return generatePropertySetter(var.getReference(),sourceType,targetType,sb);            
        }
        return null;                              
    }
    
    private static String generatePropertySetter(IReferenceableByStatementVariable ref, InvokePropertyType sourceType, InvokePropertyType targetType, StringBuilder sb) {
        if (ref instanceof ObjectProperty) {
            ObjectProperty property = (ObjectProperty)ref;
            if (property.eContainer() instanceof Component) {
                sb.append("set"+Strings.toFirstUpper(property.getName())+"(");           
            } else {
                SimpleType t = property.getType();
                sb.append(property.getName()+".set(");
            }                    
        } else if (ref instanceof CustomizeableField) {
            CustomizeableField customizeableField = (CustomizeableField)ref;
            String modelName = Strings.toFirstLower(PomNameProvider.getXmaName(customizeableField));
            sb.append(modelName+".set(");
        } else if (ref instanceof Text) {
            Text text = (Text)ref;
            String modelName = Strings.toFirstLower(PomNameProvider.getXmaName(text));
            sb.append(modelName+".set(");
        } else if (ref instanceof CheckBox) {
            CheckBox check = (CheckBox)ref;
            String modelName = Strings.toFirstLower(PomNameProvider.getXmaName(check));
            sb.append(modelName+".set(");
        } else if (ref instanceof Combo) {
            Combo combo = (Combo)ref;
            String modelName = Strings.toFirstLower(PomNameProvider.getXmaName(combo));
            sb.append(modelName+".set(");            
        } else if (ref instanceof Combo) {
            Combo combo = (Combo)ref;
            String modelName = Strings.toFirstLower(PomNameProvider.getXmaName(combo));
            sb.append(modelName+".set(");            
        } else if (ref instanceof Table) {
            Table combo = (Table)ref;
            String modelName = Strings.toFirstLower(PomNameProvider.getXmaName(combo));
            sb.append(modelName+".select(");            
        }
        return ")";                        
    }            

    private static String getInvokableTypeName(Invokeable invokeable) {
        if ( invokeable instanceof ReferencedXMAPage )
            return ((ReferencedXMAPage) invokeable).getXmaPage().getNamClass();
        if ( invokeable instanceof XmadslPage )
            return ((XmadslPage)invokeable).getName();
        if ( invokeable instanceof Component )
            return "I" + ((Component)invokeable).getXmaComponent().getNamClass();
        if ( invokeable instanceof Proxy )
            return "IComponentClient";
        return null;
    }

    private static boolean isEmbedding(Statement stmt) {
        return    stmt instanceof InvokeStatement 
               && ((InvokeStatement)stmt).getTargetPanel() != null;
    }

}
